//--------------------------------------------------------------------
// Commodore PET video
// J. Arias (2024), Public domain source
//--------------------------------------------------------------------

module PETvideo(
	input cpuck,
	input wr,
	input [7:0]d,
	input [9:0]a,
	output [7:0]q,
	output vrd,

	input pixck,
	input grset,
	output video,
	output syn,
	output hde,
	output vde,
	output frmodd
);

////////////////////////////////////////////////////////////////////
// PAL / NTSC SYNC generation stuff
// this section runs with a 500kHz clock

parameter FCLK=8000000;
localparam DIVMAX = (FCLK/500000);
localparam DIVBITS= $clog2(DIVMAX);	

parameter NLINES=525;	// 625 for PAL, 525 for NTSC

localparam NINL=(NLINES-1)/2; // lines for non-interlaced mode

// 500kHz clock for sync counters
reg [DIVBITS-1:0]ck05div=0;
wire divtc=&(ck05div | (~(DIVMAX-1))); 
always @(posedge pixck) ck05div<=divtc ? 0 : ck05div+1;
wire ck05 = ck05div[DIVBITS-1];

// PAL counter:  312 lines * 32 cycles = 9984 cycles
// NTSC counter: 262 lines * 32 cycles = 8384 cycles
localparam NCY05=NINL*32;

reg frmodd=0;	// Frame Even / Odd flip-flop
wire frmend=&(palcnt|(~(NCY05-1)));
always @(posedge ck05) if (frmend) frmodd<=~frmodd;

reg [13:0]palcnt=0;
wire [13:0]initcount = {8'h0,~frmodd,5'h0}; // skip line #0 on odd frames
always @(posedge ck05) palcnt<=frmend ? initcount : palcnt+1;

wire [8:0]rline=palcnt[13:5]; // Current raw line

wire csyn=(rline[8:3]==0);	// rlines below 8 are the composite sync
// Long pulses
wire lp=csyn&(frmodd? ((rline[2:0]==5)|(rline[2:0]==4)|((rline[2:0]==3)&( palcnt[4]))) : 
	(rline[2:0]==3)|(rline[2:0]==4)|((rline[2:0]==5)&(~palcnt[4])));
// Sync signal
wire syn=~(csyn&(palcnt[3:0]==0)|(lp&(palcnt[3:1]!=7)) | ((~csyn)&(palcnt[4:1]==0)));

localparam VBPORCH=28+6; // non displayed lines (top)
wire [8:0]frlin=rline - (8+VBPORCH);
//wire [9:0]line = {frlin,frmodd}; // visible lines, interlaced

// Data enable
reg hde=0;
always @(posedge ck05) 
	if (palcnt[4:0]==7) hde<=1; else if (palcnt[4:0]==27) hde<=0;
//	if (palcnt[4:0]==5) hde<=1; else if (palcnt[4:0]==29) hde<=0;
reg vde=0;
always @(posedge ck05)
	if (palcnt[4:0]==31) 
		if (rline==(7+VBPORCH)) vde<=1; 
		else if (rline==(7+VBPORCH+200)) vde<=0;

// pixel counter
reg[8:0]pixc;
always @(posedge pixck) pixc<= hde&vde ? pixc+1 : 0;

// Video RAM
reg [7:0]vram[0:1023];
reg [7:0]vramo;
wire vrd=hde & vde & (pixc[2:0]==0);
wire [9:0]vramar=vrd? {frlin[7:3],3'b000}+{frlin[7:3],5'b00000}+pixc[8:3] : a;

integer i;
initial for (i=0;i<1024;i=i+1) vram[i]=i&255;
reg [7:0]q;
always @(posedge pixck) vramo<=vram[vramar];
always @(posedge cpuck) if (wr) vram[a]<=d;
always @(posedge cpuck)	q<=vramo;


// Character ROM
reg [7:0]chrom[0:2047];
reg [7:0]romo;
wire [10:0]roma={grset,vramo[6:0],frlin[2:0]};
initial $readmemh("Info/romchar.list",chrom);
always @(posedge pixck) romo<=chrom[roma];

// Shift
reg d7d;
always @(posedge pixck) d7d<=vramo[7];
reg [7:0]sh;
wire load=hde & vde & (pixc[2:0]==2);

always @(posedge pixck) sh <= load ? (d7d ? ~romo : romo) : {sh[6:0],1'b0};
assign video=sh[7];

endmodule

