//////////////////////// keyboard PS2 -> PET //////////////////////
//
// J. Arias (2024), Public domain source
//
// PET keyboard is a 10-row, 8-column, switching matrix (80 keys)
// that matrix is simulated by an 80-bit RAM where a bit
// is written with 0 if a PS2 key is pressed.
// This module has basically 3 parts:
//  1 - PS2 serial to parallel receiver
//  2 - Mapping ROM (scancode to matrix address)
//  3 - Matrix RAM

module keybps2(
input clk,			// CPU clock
input kclk,			// PS2 clock
input kdat,			// PS2 data
input [3:0]a,		// row address 
output [7:0]row,	// output row 
output resetk		// Reset key pressed
);

////////////////////////////////////////////////////
// PS2 serial receiver
////////////////////////////////////////////////////

reg [10:0]kreg=11'h7FF;	// shift register, 11-bit
reg krck=1;				// sampled keyboard clock
always @(posedge clk) krck<=kclk;	

reg kock=1;			// old clock sample (for edge detection)
wire newsc;			// new scancode received
wire [7:0]scancode;	
reg rel=0;					// key released (if last scancode was 0xF0)
assign newsc=~kreg[0];		// zero start bit => scancode received
assign scancode=kreg[8:1];

// Timeout counter (4.096ms)
wire fall=(~krck)&kock;	// falling clock
reg [11:0]toutc=0;		// timeout counter
always @(posedge clk) toutc<= fall? 0: toutc+1;
wire tout=(toutc==12'hFFF)&(~fall);

// shift reg.gets filled with ones after a key is received
// or when a timeout happens
always @(posedge clk) begin
	kock<=krck;
	kreg<=(newsc|tout) ?11'h7FF : ( fall ? {kdat,kreg[10:1]} : kreg);
	if (newsc) rel<=(scancode==8'hF0);
end

// Special keys
reg resetk;					// Reset key pressed (F12)
always @(posedge clk) 
  if ((scancode==8'h07)&newsc) resetk<=~rel;
  
reg fsh=0;			// Forced shift (for 2 cursor keys)
always @(posedge clk) 
  if (((scancode==8'h6b)|(scancode==8'h75))&newsc) fsh<=~rel;


////////////////////////////////////////////////////
// scancode to keyboard array mapping (ROM)
////////////////////////////////////////////////////

reg [6:0]keypos;	// bit address (not a register)
always@*
 casex (scancode)

// Cursors
 8'h6B: keypos<=7'd5;	// Left arrow  (crsr <>)+shift
 8'h74:	keypos<=7'd5;	// Right arrow (crsr <>)
 8'h75: keypos<=7'd44;	// Up arrow    (crsr ^v)+shift
 8'h72: keypos<=7'd44;	// Down arrow  (crsr ^v)

 8'h77: keypos<=7'd76;		// Pause (Break)
// 8'h01: keypos<=7'h03;	// F9
// 8'h0B: keypos<=7'h04;	// F6
 8'h04: keypos<=7'd13;	// F3 
 8'h5A: keypos<=7'd28;	// Enter

// 8'h78: keypos<=7'h09; // F11
// 8'h0A: keypos<=7'h0B; // F8
// 8'h03: keypos<=7'h0C; // F5
 8'h05: keypos<=7'd64;	// F1 (off/rvs)
 8'h06: keypos<=7'd72;	// F2 
// 8'h09: keypos<=7'h0F; // F10

 8'h54: keypos<=7'd46;	// [
// 8'h77: keypos<=7'h12;	// RET (Alt)
 8'h5B: keypos<=7'd20;	// ]
 8'h0C: keypos<=7'd62;	// F4
 8'h12: keypos<=7'd54;  // Shifts
 8'h59: keypos<=7'd48;	// Shifts
 8'h0E: keypos<=7'd36;	// \ 
// 8'h14: keypos<=7'h17;	// Ctrl (both L and R)

 8'h55: keypos<=7'd30;	// ¡ (@)
 8'h4A: keypos<=7'd3;	// -= 
 //8'h5D: keypos<=7'd8;	// |
 8'h4D: keypos<=7'd38;	// P
 8'h11: keypos<=7'd77;	// ;
 8'h27: keypos<=7'd22;	// : (windows)
 8'h49: keypos<=7'd51;	// , 
 8'h4E: keypos<=7'd70;	// /


 8'h45: keypos<=7'd11;	// 0
 8'h46: keypos<=7'd75;	// 9
 8'h44: keypos<=7'd45;	// O
 8'h43: keypos<=7'd37;	// I
 8'h4B: keypos<=7'd29;	// L
 8'h42: keypos<=7'd21;	// K
 8'h3A: keypos<=7'd67;	// M
 8'h41: keypos<=7'd59;	// .

 8'h3E: keypos<=7'd2;	// 8
 8'h3D: keypos<=7'd10;	// 7
 8'h3C: keypos<=7'd43;	// U
 8'h35: keypos<=7'd35;	// Y
 8'h33: keypos<=7'd19;	// H
 8'h3B: keypos<=7'd27;	// J
 8'h31: keypos<=7'd58;	// N
 8'h29: keypos<=7'd66;	// space

 8'h36: keypos<=7'd74;	// 6
 8'h2E: keypos<=7'd1;	// 5
 8'h2D: keypos<=7'd34;	// R
 8'h2C: keypos<=7'd42;	// T
 8'h34: keypos<=7'd26;	// G
 8'h2B: keypos<=7'd18;	// F
 8'h32: keypos<=7'd50;	// B
 8'h2A: keypos<=7'd57;	// V

 8'h25: keypos<=7'd9;	// 4
 8'h26: keypos<=7'd73;	// 3
 8'h24: keypos<=7'd41;	// E
 8'h1D: keypos<=7'd33;	// W
 8'h1B: keypos<=7'd17;	// S
 8'h23: keypos<=7'd25;	// D
 8'h21: keypos<=7'd49;	// C
 8'h22: keypos<=7'd65;	// X

 8'h16: keypos<=7'd8;	// 1
 8'h1E: keypos<=7'd0;	// 2
 8'h76: keypos<=7'd16;	// Esc
 8'h15: keypos<=7'd40;	// Q
 8'h0D: keypos<=7'd32;	// Tab
 8'h1C: keypos<=7'd24;	// A
// 8'h58: keypos<=7'd;	// CapsLock 
 8'h1A: keypos<=7'd56;	// Z
 
 8'h6C: keypos<=7'd68;	// Home
 8'h66: keypos<=7'd39;	// Del
 default: keypos<=7'b111_xxxx;
endcase

////////////////////////////////////////////////////
//   Matrix RAM 
////////////////////////////////////////////////////

reg [7:0]ka[0:9]; 	// 10 8-bit words 
wire wrk;			// write to keyboard RAM
assign wrk=newsc&(~(&keypos[6:4]));
//assign wrk=newsc&(~(keypos[6]&keypos[4]));
always @(posedge clk) begin
	// single bit write
	if (wrk)ka[keypos[6:3]][keypos[2:0]] <= rel;
end
// 8-bit read
reg [7:0]rowm; // sync => BRAM
always @(posedge clk) rowm<=ka[a];
// output include forced shift
assign row={rowm[7],rowm[6]&(~(fsh&(a==6))),rowm[5:0]};

// start with all bits as 1 (keys not pressed)
integer i;
initial begin
  for (i=0;i<=9;i=i+1) ka[i] = 8'hff;
end

endmodule

