///////////////////////////////////////////////
//            LOGIC ANALYZER
//
// Teresa Díez & Jesús Arias (2020-2024)
//
// Public domain sources
///////////////////////////////////////////////

module lanal (
	input clk,		// clock, sample on rising edges
	input rxd,		// Serial input (commands)
	input [31:0]I,	// Signals to sample
	output txd,		// Serial output (captured data)
	output [3:0]st	// Status (for debug / LEDs)
);

parameter BUFFER_SIZE=8192;	// Must be a power of two
parameter FCLK=96000000;	// Clock frequency (Hz)
parameter BAUD=1500000;     // Serial data rate (bits/s)
localparam MW = $clog2(BUFFER_SIZE)-3; // 2^MW 64-bit longwords (MW>=8)
localparam UDIV = (FCLK+BAUD/2)/BAUD;  // UART divider

//////////////////////////////////////////////////////////////////////////////////
// Datapath
//////////////////////////////////////////////////////////////////////////////////

// UART
wire [7:0] uart_do;	// RX data
wire [7:0] uart_di;	// TX data
wire rxrdy;			// RX data ready (active for just one clock cycle)
wire txrdy;			// TX ready
wire txwr;			// TX write (starts TX)

lanal_UART_core #(.DIVIDER(UDIV)) uart0( 
	.clk(clk), .txd(txd), .rxd(rxd), .txrdy(txrdy),.d(uart_di),
	.wr(txwr), .q(uart_do), .rxvalid(rxrdy), .nstop(1'b0) );
	
// Input signal MASK register
reg [31:0]msk=32'hffffffff;	
wire wrmsk0,wrmsk1,wrmsk2,wrmsk3;	// write lanes to Input Mask

always @(posedge clk)
begin
	if (wrmsk0) msk[ 7: 0]<=uart_do;
	if (wrmsk1) msk[15: 8]<=uart_do;
	if (wrmsk2) msk[23:16]<=uart_do;
	if (wrmsk3) msk[31:24]<=uart_do;
end

// Trigger Mask register
reg [31:0]tmsk=32'hffffffff;	
wire wrtmsk0,wrtmsk1,wrtmsk2,wrtmsk3;	// write lanes to trigger Mask

always @(posedge clk)
begin
	if (wrtmsk0) tmsk[ 7: 0]<=uart_do;
	if (wrtmsk1) tmsk[15: 8]<=uart_do;
	if (wrtmsk2) tmsk[23:16]<=uart_do;
	if (wrtmsk3) tmsk[31:24]<=uart_do;
end

reg [31:0]ffI;	// registered inputs
always @(posedge clk) ffI<=I;

wire [31:0]rId; // Inputs after mask
reg  [31:0]rIq;	// the same one cycle before
assign rId=ffI&msk;
always @(posedge clk) rIq<=rId;

wire evento;	// Data change event
wire trigger;
wire [31:0] Ixor;
assign Ixor=rId^rIq;
assign evento=(Ixor!=0);
assign trigger=((Ixor&tmsk)!=0);
reg [31:0]tiempo=0;	// Time counter
wire restim;		// Time reset
always @(posedge clk) tiempo <= restim ? 0 : (tiempo+1);

reg [7:0]timeout=0;	// Timeout register
wire wrtout;		// write to timeout reg.
always @(posedge clk) if (wrtout) timeout<=uart_do;
wire tout;			// Timeout flag

`ifdef SIMULATION
	assign tout=(timeout==tiempo[11:4]);	// short timeout for simulation
`else
	assign tout=(timeout==tiempo[31:24]);
`endif

reg [MW:0]dir=0;	// RAM Address counter
wire resdir;		// reset address
wire incdire;		// increment address
wire maxmem;		// memory full
always @(posedge clk) dir <= resdir ? 0 : dir + incdire;
assign maxmem=dir[MW];

// Synchronous RAM memory
wire wrmem;			// write enable
reg [63:0]memdat;	// output register

reg [63:0]mem[0:(1<<MW)-1];
always @(posedge clk) begin
	if (wrmem) mem[dir[MW-1:0]]<={tiempo,rId};
	memdat<=mem[dir[MW-1:0]];
end

reg [MW:0]maxdir;	// Maximum address register 
wire wrmaxdir;		// write to max address register 
always @(posedge clk) if (wrmaxdir) maxdir<=dir;
wire direq;			// address counter matches max reg.
assign direq=(dir==maxdir);

reg [2:0]bytesel;	// byte counter for ouput multiplexing
wire incbyte;		// increment byte counter
wire resbyte;		// reset byte counter
wire bytetc;		// terminal count (byte #7)
always @(posedge clk) bytesel <= resbyte ? 0 : bytesel + incbyte;
assign bytetc=&bytesel;

// output data mux
//wire [7:0]membyte = memdat[{bytesel,3'b111}:{bytesel,3'b000}];
wire [7:0]mb[0:7];
assign mb[0]=memdat[7:0];
assign mb[1]=memdat[15:8];
assign mb[2]=memdat[23:16];
assign mb[3]=memdat[31:24];
assign mb[4]=memdat[39:32];
assign mb[5]=memdat[47:40];
assign mb[6]=memdat[55:48];
assign mb[7]=memdat[63:56];
wire [7:0]membyte = mb[bytesel];

wire txdirh;		// address byte selection (1=MSB), (0=LSB) 
wire txdato;		// data or address selection (1=data) (0=address)

assign uart_di=txdato ? membyte : (txdirh ? {{MW-8{1'b0}},maxdir[MW:8]} : maxdir[7:0]);

wire cmdfd;			// command 0xFD on UART data
assign cmdfd=(uart_do==8'hfd);
wire cmdfe;			// command 0xFE on UART data
assign cmdfe=(uart_do==8'hfe);

//////////////////////////////////////////////////////////////////////////////////
// State-Machine Controller
//////////////////////////////////////////////////////////////////////////////////

// State definition:
// <8 : operation
parameter IDLE=		0;
parameter TRIGGER=	1;
parameter CAPTURA=	2;
parameter TXDIRL=	3;
parameter TXDIRH=	4;
parameter DDUMP=	5;
// >=8 : config (enable-register write)
parameter MSK0=		8;
parameter MSK1=		9;
parameter MSK2=		10;
parameter MSK3=		11;
parameter TMSK0=	12;
parameter TMSK1=	13;
parameter TMSK2=	14;
parameter TMSK3=	15;

reg  [3:0]st=IDLE;	// current state

// Next state logic
always @(posedge clk)
begin
	if (rxrdy & cmdfd & (~st[3])) st<=IDLE;	// reset command
	else
		case (st)
		IDLE:		if (rxrdy) st<= cmdfe ? MSK0 : TRIGGER;
		TRIGGER:	if (tout) st<=TXDIRL; else if (trigger) st<=CAPTURA;
		CAPTURA:	if (maxmem | tout) st<=TXDIRL;
		TXDIRL:		if (txrdy) st<=TXDIRH;
		TXDIRH:		if (txrdy) st<=DDUMP;
		DDUMP:		if (direq) st<=IDLE;

		MSK0:		if (rxrdy) st<=MSK1;
		MSK1:		if (rxrdy) st<=MSK2;
		MSK2:		if (rxrdy) st<=MSK3;
		MSK3:		if (rxrdy) st<=TMSK0;
		TMSK0:		if (rxrdy) st<=TMSK1;
		TMSK1:		if (rxrdy) st<=TMSK2;
		TMSK2:		if (rxrdy) st<=TMSK3;
		TMSK3:		if (rxrdy) st<=IDLE;
		default:	st<=IDLE;
		endcase
end

// control outputs
assign wrmsk0=(st==MSK0) & rxrdy;
assign wrmsk1=(st==MSK1) & rxrdy;
assign wrmsk2=(st==MSK2) & rxrdy;
assign wrmsk3=(st==MSK3) & rxrdy;

assign wrtmsk0=(st==TMSK0) & rxrdy;
assign wrtmsk1=(st==TMSK1) & rxrdy;
assign wrtmsk2=(st==TMSK2) & rxrdy;
assign wrtmsk3=(st==TMSK3) & rxrdy;

assign wrtout= (st==IDLE) & rxrdy & (~cmdfe);
assign restim=((st==IDLE) & rxrdy & (~cmdfe)) | ((st==TRIGGER) & trigger);

assign resdir=((st==IDLE) & (~rxrdy)) | (st==TXDIRH);
assign wrmem= ((st==IDLE) & rxrdy & (~cmdfe)) | ((st==TRIGGER) & trigger) | ((st==CAPTURA) & evento);
assign wrmaxdir= ((st==TRIGGER) & tout) | ((st==CAPTURA) & (tout | maxmem));
assign incbyte= ((st==DDUMP) & txrdy);
assign resbyte=(st==IDLE);

assign incdire= wrmem | ((st==DDUMP) & txrdy & bytetc);

assign txwr=txrdy&((st==TXDIRL) | (st==TXDIRH) |(st==DDUMP));
assign txdirh=(st==TXDIRH);
assign txdato=(st==DDUMP);


endmodule

///////////////////////////////////////////////////////////////////////////////
// Simple UART: data formats: 8N1, 8N2. Fixed baud rate (selected by DIVIDER)
// by J. Arias
///////////////////////////////////////////////////////////////////////////////
module lanal_UART_core (
	input	clk,		// System clock
	input	[7:0]d,		// Input bus for TX
	input	wr,			// write strobe: starts TX
	output	[7:0]q,		// Output bus for RX
	output	txrdy,		// TX ready flag
	output	rxvalid,	// RX valid data flag
	input   nstop,		// Set to 1 for two stop bits
	input 	rxd,		// serial data input
	output 	txd			// serial data output
);
parameter DIVIDER= 217;	// BAUD = f_clk / DIVIDER
localparam DBITS = $clog2(DIVIDER);	// number of bits for clock counters

///////////////////// UART TX ////////////////////
reg [DBITS-1:0]txdiv=0;
wire txdivmax=(txdiv==(DIVIDER-1));
always @(posedge clk) 
	if (wr | txdivmax) txdiv<=0; // Reset if max or on writes
	else txdiv<=txdiv+1;

reg [8:0]txsh=10'h1FF;	// 9-bit shift register
always @(posedge clk)
	if (wr) begin 
		`ifdef SIMULATION
	        $write ("%c",d); $fflush ( );
		`endif
		 txsh<={d,1'b0}; // Load data plus START bit
	end	else if (txdivmax) txsh<={1'b1,txsh[8:1]}; // shift 1'b1 in for future STOP bit
assign txd = txsh[0];

reg [3:0]txbitcnt=0;	// Bit counter (0000 means idle)
wire utxbusy=|txbitcnt;
always @(posedge clk) 
	if (wr) txbitcnt<={3'b101,nstop}; // set to eleven for 2 STOP bits, ten for 1 STOP bit
	else if (utxbusy&txdivmax) txbitcnt<=txbitcnt-1;
	
assign txrdy=~utxbusy;
	
///////////////////// UART RX (no buffer) ////////////////////
reg [1:0]rxreg=0; // two samples of RXD
always @(posedge clk) rxreg<={rxreg[0],rxd};
reg [DBITS-1:0]rxdiv=0;
always @(posedge clk)
	if ((rxreg[1]^rxreg[0])|(rxdiv==(DIVIDER-1))) rxdiv<=0; // Reset if max or on any RXD edge
	else rxdiv<=rxdiv+1;
wire rxsample = (rxdiv==(DIVIDER/2-1)); // sample at the middle of the bit

reg [9:0]urxsh=10'h3FF; // 10-bit shift register
always @(posedge clk)
	if (~urxsh[0]) urxsh<=10'h3FF;  // if START bit at LSB
	else if (rxsample) urxsh<={rxreg[1],urxsh[9:1]};

assign q = urxsh[8:1];
assign rxvalid = ~urxsh[0];	// START bit at LSB

endmodule


