//--------------------------------------------------------------------
// Commodore PET 2001-4K (or 8k if big FPGA)
// J. Arias (2024), Public domain source
//--------------------------------------------------------------------

`include "PETscii.v"
`include "cpu6502.v"
`include "keyb.v"
`include "pia6520.v"
`include "via6522.v"

`include "tapefifo.v"

//`include "lanal.v"

module SYSTEM(
	input clk,		// 48 MHz
	inout video,
	output syn,
	
	output [2:0]ant,	// antenna DAC

	// SPI bus (To Flash and SD card)
	output sck,
	output mosi,
	output ssb,		// Flash /CS
	input  miso,
	
	// PS2
	input kclk,
	input kdat,
	
	output audio

`ifdef XRAM
	,inout [15:0]xd,
	output [15:0]xa,
	output xble,
	output xbhe,
	output xoe,
	output xwe
`endif
	// debug 
	, input rxd
	, output txd
	, output [2:0]debug
);

`ifdef ICE1K
	parameter RAMSZ=4096;
`else
	parameter RAMSZ=8192;
`endif

//////////////
// Clocks
// Prescaler 1/6 = 8 MHz
reg [1:0]pre1=0;
always @(posedge clk) pre1<= pre1[1] ? 0 : pre1+1;
reg pixck=0; // pixel clock
always @(posedge pre1[1]) pixck<=~pixck; 
// Prescaler 1/8 interleaved with video reads
reg [2:0]pre2=0; // VRD => reset on next cycle
always @(posedge pixck) pre2<= vrd ? 0 : pre2+1;
// CPU clock = 1 MHz
wire cclk = (pre2!=6)|vrd|pixck;

//////////////
// Reset
`ifdef SIM
 reg [3:0]rstcnt=2;
 always @(posedge hde) rstcnt<= |rstcnt ? rstcnt-1 : 0;
 reg reset=1;
 always @(posedge cclk) reset<= |rstcnt;
`else
 reg [3:0]rstcnt=15;
 always @(posedge vde) rstcnt<= |rstcnt ? rstcnt-1 : 0;
 reg reset=1;
 always @(posedge cclk) reset<= |rstcnt;
`endif

/////////////////////
// Video controller

wire vid,syn,vde,vrd,hde;
wire vramwe = we & (ca[15:12]==4'b1000) &(~ca[11]);  // write to VRAM
wire [7:0]vramo;	// data read from VRAM

PETvideo video0(
	.pixck(pixck), .cpuck(cclk), .video(vid), .syn(syn), .vde(vde), .hde(hde),
	.wr(vramwe), .d(cdo), .a(ca[9:0]), .q(vramo), .vrd(vrd), .grset(grbuss)
);
// Composite video signal
`ifndef VIDEO2PIN
  assign video= (syn&(~vid))? 1'bz: syn;  
`else
  assign video=vid;
`endif
// Antenna DAC (thermometer code)
`ifdef RF
	assign ant[0]=clk;				// 330ohm  (amp=17%)
	assign ant[1]=clk & (~vid);		// 100ohm  (amp=74%)
	assign ant[2]=clk & (~syn);		// 220ohm  (amp=100%)
`endif
reg vramdrd=0;	// VRAM select, delayed (for cdi mux)
always @(posedge cclk) vramdrd<= (~we) & (ca[15:12]==4'b1000) &(~ca[11]);

/////////////////////
//  CPU
wire [7:0]cdo;	// CPU data output
wire [7:0]cdi;	// CPU data input
wire [15:0]ca;  // CPU address
wire we;		// write enable
wire irq;		// Interrupt input

cpu6502 cpu6502( .AB(ca),.DO(cdo),.WE(we),.DI(cdi),.clk(cclk),.reset(reset|resetk),
	.IRQ(irq), .NMI(1'b0), .RDY(1'b1) );

// cdi mux
assign cdi =
	(ramdrd  ?  ramo : 8'hFF) &
	(vramdrd ? vramo : 8'hFF) &
	(romdrd  ?  romo : 8'hFF) &
	(iodrd   ?   ioo : 8'hFF) &
	(snard   ?  snao : 8'hFF) ;

/////////////////////
// RAM memory

`ifdef XRAM // 32K PET
	reg ramdrd=0;	// RAM select, delayed (for cdi mux)
	always @(posedge cclk) ramdrd<=(~we) & (~ca[15]);
/*	assign xa = {1'b0,ca[15:1]};
	assign xble=ca[0];
	assign xbhe=~ca[0];
	assign xoe=~((~ca[15])&(~we));
	assign xwe=~((~ca[15])& we & (~cclk));
	assign xd = xoe ? {cdo,cdo} : 16'hZZZZ;
	reg [7:0]ramo;							// data read from RAM
	always @(posedge cclk) ramo<= ca[0] ? xd[15:8] : xd[7:0];
*/
	assign xa = ca[15:0];
	assign xble=1'b0;
	assign xbhe=1'b1;
	assign xoe=~((~ca[15])&(~we));
	assign xwe=~((~ca[15])& we & (~cclk));
	assign xd = xoe ? {8'hXX,cdo} : 16'hZZZZ;
	reg [7:0]ramo;							// data read from RAM
	always @(posedge cclk) ramo<= xd[7:0];
`else
	localparam RAMAW= $clog2(RAMSZ);

	reg [7:0]ram[0:RAMSZ-1];
	wire ramwe = we & (ca[15:RAMAW]==0);	// write to RAM
	reg [7:0]ramo;							// data read from RAM
	always @(posedge cclk) begin
		ramo<=ram[ca[RAMAW-1:0]];
		if (ramwe) ram[ca[RAMAW-1:0]]<=cdo;
	end
	reg ramdrd=0;	// RAM select, delayed (for cdi mux)
	always @(posedge cclk) ramdrd<=(~we) & (ca[15:RAMAW]==0);
`endif
/////////////////////
// ROM memory
wire [7:0]romo;	// data read from ROM
spirom rombank( .clk(clk), .cclk(cclk), .sck(sck), .mosi(mosi), .miso(miso), .csb(ssb),
		.addr(ca[13:0]), .data_out(romo) );

reg romdrd=0;	// ROM select, delayed (for cdi mux)
always @(posedge cclk) romdrd<=(~we) & ca[15] & ca[14] & (~iosel);

/////////////////////
// SNA ROM/RAM memory
	reg [7:0]sna[0:511];
	initial $readmemh("snarom.hex",sna);
	wire snawe = we & (ca[15:12]==4'h8)&ca[11]&ca[8];	// write to RAM
	reg [7:0]snao;							// data read from RAM
	always @(posedge cclk) begin
		snao<=sna[ca[8:0]];
		if (snawe) sna[ca[8:0]]<=cdo;
	end
	reg snard=0;	// Mem select, delayed (for cdi mux)
	always @(posedge cclk) snard<=(~we) & (ca[15:12]==4'h8)&(ca[11]);

/////////////////////
// I/O
wire iosel = (ca[15:11]==5'b11101); // 0xE8xx
reg iodrd=0;	// I/O select, delayed (for cdi mux)
always @(posedge cclk) iodrd<=(~we) & iosel;
reg [7:0]ioo;	// data read from I/O


`ifdef ICE1K	// 1280 LCs

		always @(posedge cclk)
		  casex (ca[7:0])
		  8'bx000xxx0:  ioo<=rxbyte;
		  8'bx000xxx1:  ioo<={rxdav,7'b0};
		  // PIA1
		  8'bxxx1xx00:	ioo<={4'b1111,pia1a};
		  8'bxxx1xx01:  ioo<=8'b00000100;
		  8'bxxx1xx10:	ioo<=pia1b;
		  8'bxxx1xx11:	ioo<={piairq,6'b000010,p1ie};
		  // VIA
		  //8'bx100xxxx:	ioo<=viado;
		  default:	ioo<=8'hff;
		  endcase

		/////////////////////////////////////
		// minimal PIA1 (keyboard)
		reg [3:0]pia1a=0;	// Port A outputs (keyboard row select)
		always @(posedge cclk) begin
			if (we & iosel & (ca[6:4]==1) & (ca[1:0]==0)) pia1a<=cdo[3:0];
		end

		reg p1if=0;		// Interrupt flag (CB1 rising edge = Vert_DE falling edge)
		reg vded=0;		// delayed VDE 
		always @(posedge cclk) begin
			vded<=vde;
			p1if<=(~vde)&vded ? 1 : (iosel&(ca[6:4]==1)&(ca[1:0]==2)&(~we) ? 0 : p1if);
		end
		wire piairq=p1if;	// PIA1 interrupt (CB1)
		assign irq = piairq;
		/////////////////////////////////////
		// Keyboard matrix emulation from PS2
		wire [7:0]pia1b;
		keybps2 ps2(
			.clk(cclk), .kclk(kclk), .kdat(kdat),	// PS2
			.a(pia1a[3:0]), .row(pia1b),			// PIA
			.resetk(resetk)					// reset key (F12)
		);
		wire grbuss=0;
		
		/////////////////////////////////////////////////
		// UART Receiver for SNAPSHOT loading
		parameter FCLK= 8000000;
		parameter BAUD= 115200;		// UART baudrate
		localparam DIVIDER=(FCLK+BAUD/2)/BAUD;
		localparam DBITS = $clog2(DIVIDER);     // number of bits for clock counters

		reg [1:0]rxreg=0; // two samples of RXD
		always @(posedge pixck) rxreg<={rxreg[0],rxd};
		wire rxchange = rxreg[1]^rxreg[0];
		reg [DBITS-1:0]rxdiv=0;
		wire rxdivtc = &(rxdiv|~(DIVIDER-1));
		always @(posedge pixck)
			if (rxchange | rxdivtc ) rxdiv<=0; // Reset if max or on any RXD edge
			else rxdiv<=rxdiv+1;
		wire rxsample = (rxdiv==(DIVIDER/2-1)) & (~rxchange); // sample at the middle of the bit

		reg [8:0]rxsh=9'h1FF; // 10-bit shift register
		reg [7:0]rxbyte;
		reg rxdav = 0;	
		always @(posedge pixck) begin
			if (~rxsh[0]) begin
				rxsh<=9'h1FF;  // if START bit at LSB set to all ones and copy results
				rxbyte<=rxsh[8:1];
			end	else if (rxsample) rxsh<={rxreg[1],rxsh[8:1]};
			rxdav<= (~rxsh[0]) ? 1 : (iosel&(ca[6:4]==0)&(~ca[0]) ? 0 :rxdav);
		end
		
`else	// 7680 LCs

		always @(posedge cclk)
		  casex (ca[6:4])
		  // UART
		  3'b000:   ioo<=ca[0] ? {rxav,7'b000000} : rxdat;
		  // PIAs
		  3'bxx1:	ioo<=pia1do;
		  3'bx10:	ioo<=pia2do;
		  // VIA
		  3'b100:	ioo<=viado;
		  default:	ioo<=8'hff;
		  endcase

		/////////////////////////////////////
		// PIA1 6520
		wire pia1sel=iosel&ca[4];
		wire [7:0]pia1do;
		wire pia1irq;
		wire cass_motor_n;
		wire cass_read;
		pia6520 pia1(
			.clk(cclk), .reset(reset), .we(we), .strobe(pia1sel),
			.data_in(cdo), .data_out(pia1do), .addr(ca[1:0]),
			.irq(pia1irq),
			.porta_out(pia1a), .porta_in(8'hC0), .portb_in(pia1b), .cb1_in(vde),
			.cb2_out(cass_motor_n), .cb2_in(cass_motor_n),
			.ca1_in(cass_read)
		);
		/////////////////////////////////////
		// Keyboard matrix emulation from PS2
		wire [7:0]pia1a;
		wire [7:0]pia1b;
		keybps2 ps2(
			.clk(cclk), .kclk(kclk), .kdat(kdat),
			.a(pia1a[3:0]), .row(pia1b),
			.resetk(resetk)					// reset key (F12)
		);

		/////////////////////////////////////
		// PIA2 6520
/*
		wire pia2sel=iosel&ca[5];
		wire [7:0]pia2do;
		wire pia2irq;

		pia6520 pia2(
			.clk(cclk), .reset(reset), .we(we), .strobe(pia2sel),
			.data_in(cdo), .data_out(pia2do), .addr(ca[1:0]),
			.ca1_in(1'b1), .cb1_in(1'b1),
			.irq(pia2irq)
		);
*/
		wire pia2irq=0;
		wire [7:0]pia2do=8'hXX;
		/////////////////////////////////////
		// VIA 6522
		wire viasel=iosel&ca[6];
		wire [7:0]viado;
		wire viairq;
		wire grbuss;
		wire [7:0]viapbin={2'b11,vde,5'b11111};
		wire [7:0]viapbo;
		wire cass_write=viapbo[3];
		wire audio;

		via6522 via(
			.clk(cclk), .reset(reset), .slow_clock(1'b1), .we(we), .strobe(viasel),
			.data_in(cdo), .data_out(viado), .addr(ca[3:0]),
			.irq(viairq),
			.ca1_in(1'b1),
			.ca2_out(grbuss), .ca2_in(grbuss), .portb_in(viapbin), .portb_out(viapbo),
			.cb2_out(audio), .cb2_in(audio),
			.cb1_in(cass_read)
		);
		assign irq = pia1irq | pia2irq | viairq;

	//////////////////////////////////////////////
	// Cassete emulation
	tapefifo #(.SIZE(512)) tpf0 (.clk(pixck), .rxd(rxd), .txd(txd) ,
					.cass_read(cass_read), .cass_write(cass_write),
					.rxbyte(rxbyte), .rxdav(rxdav) );
	assign debug[0]=cass_read;
	assign debug[1]=cass_write;

	///////////////////////////////////////////////
	// UART RX interface for SNAPSHOT loading
	wire [7:0]rxbyte;
	wire rxdav;
	reg [7:0]rxdat;
	reg rxav=0;
	always @(posedge pixck) begin
		if (rxdav) rxdat<=rxbyte;
		rxav<= rxdav ? 1 : (iosel&(~ca[6])&(~ca[4])&(~ca[4])&(~ca[0]) ? 0 : rxav);
	end

`endif

endmodule

//----------------------------------------------------------------------------
//-- SPI ROM memory
//----------------------------------------------------------------------------

module spirom (
    input  clk,
    input  cclk,
    //input  rd,
    
    output sck,
    output csb,
    output mosi,
    input  miso,
    
    input  [13:0] addr,      
    output [7:0] data_out   
);

reg dcclk=0;
always @(posedge clk) dcclk<=cclk;
wire cclkedge = (cclk) & (~dcclk);

reg [5:0]bitcnt=6'd63;

always @(posedge clk) 
	if (cclkedge) bitcnt<=0;
		else if ((bitcnt!=6'd63)) bitcnt<=bitcnt+1;

reg csb=1;
always @(posedge clk) 
	if (cclkedge) csb<=0;
	//else if ((bitcnt==6'd15)&(~rd)) csb<=1;
	else if (bitcnt>=39) csb<=1;
assign sck = (~csb)&(~clk);	

reg [15:0]sh=0;

always @(posedge clk )
	if (cclkedge) sh<=16'h0302;
	else begin
		if (bitcnt==6'd15) sh<={2'b11,addr};
		else if (~csb) sh<={sh[15:0],miso};
	end
	
assign mosi=sh[15];
reg [7:0]data_out;
always @(posedge cclk) data_out<=sh[7:0];
//assign data_out=sh[7:0];
	
endmodule



