//--------------------------------------------------------------------
// Serial port to Commodore Cassette tape adapter
// J. Arias (2024), Public domain source
//
// - Data format: .TAP files (byte * 8 = square cycle lenght in us)
// - FIFO buffer with flow control (XON / XOFF)
// 
//--------------------------------------------------------------------

module tapefifo(
	input  clk,
	input  rxd,
	output txd,
	input  cass_write,
	output cass_read,
	output [7:0]rxbyte,
	output rxdav
);
parameter SIZE= 512;		// FIFO size (1 BRAM)
parameter FCLK= 8000000;	// pixel clock: 8MHz, do not change
parameter BAUD= 115200;		// UART baudrate
localparam DIVIDER=(FCLK+BAUD/2)/BAUD;
localparam DBITS = $clog2(DIVIDER);     // number of bits for clock counters

/////////////////////////////////////////////////
// UART Receiver
reg [1:0]rxreg=0; // two samples of RXD
always @(posedge clk) rxreg<={rxreg[0],rxd};
wire rxchange = rxreg[1]^rxreg[0];
reg [DBITS-1:0]rxdiv=0;
wire rxdivtc = &(rxdiv|~(DIVIDER-1));
always @(posedge clk)
    if (rxchange | rxdivtc ) rxdiv<=0; // Reset if max or on any RXD edge
    else rxdiv<=rxdiv+1;
wire rxsample = (rxdiv==(DIVIDER/2-1)) & (~rxchange); // sample at the middle of the bit

reg [9:0]rxsh=10'h3FF; // 10-bit shift register
wire [7:0]rxbyte=rxsh[8:1];
wire rxdav = ~rxsh[0];	// data available if START bit at LSB
always @(posedge clk) 
    if (rxdav) rxsh<=10'h3FF;  // if START bit at LSB set to all ones and copy results
    else if (rxsample) rxsh<={rxreg[1],rxsh[9:1]};

/////////////////////////////////////////////////
// FIFO
localparam AW=$clog2(SIZE); 
reg [7:0]mem[0:SIZE-1];
reg [AW-1:0]wrp=0;
reg [AW-1:0]rdp=0;
reg [7:0]memout=0;

always @(posedge clk) begin
	if (rxdav) begin
		mem[wrp]<=rxbyte;
		wrp<=wrp+1;
	end
	memout<=mem[rdp];	
	if (rdpulse) rdp<=rdp+1;
end
	
wire [AW-1:0]nfifo = wrp-rdp; // Number of bytes in the FIFO
wire dav=|nfifo;	// data available

// XON / XOFF Thresholds
`ifdef SIM
	wire uthr = (nfifo == 4) & (~xoff);
	wire lthr = (nfifo == 2) & xoff;
`else
	wire uthr = (nfifo == SIZE/2+16) & (~xoff);
	wire lthr = (nfifo == SIZE/2-16) & xoff;
`endif
reg xoff=0;	// XOFF state
always @(posedge clk) xoff<= uthr ? 1: (lthr ? 0 : xoff);

///////////////////////////////////////////////////////////
// FIFO data digestion: .TAP file decoder (Cassette input)
reg ddav=0; // delayed data availabe (sync mem)
always @(posedge clk) ddav<=dav;

reg cass_q=0;
reg [12:0]cascnt=0;
wire casstc1 = (cascnt==0);
always @(posedge clk) begin
	cascnt<= ddav? (casstc1 ? {memout[7:0],5'b11111} : cascnt-1) : 13'h001f;
	cass_q<= ddav? (cass_q^casstc1): 0;
end
wire rdpulse= (cascnt==2)& cass_q &ddav;	
assign cass_read = ddav & cass_q;

////////////////////////////////////////////////////////
// Cassette output: .TAP encoding (time bw rising edges)
reg ocasw=0;
always @(posedge clk) ocasw<=cass_write;
wire uwr = ((~ocasw) & cass_write); // rising edge: UART write
reg [13:0]pl=0; // pulse length
always @(posedge clk) pl<= uwr ? 0 : pl+1;

/////////////////////////////////////////////////
// UART Transmitter
reg [DBITS-1:0]txdiv=0;
wire txdivmax=&(txdiv|~(DIVIDER-1));
wire txwr = uthr | lthr | uwr;
wire [7:0]txdata= (uthr ? 8'd19: 0) |
				  (lthr ? 8'd17: 0) |
				  (uwr  ? pl[13:6] :0 );

always @(posedge clk) 
    if (txwr | txdivmax) txdiv<=0; // Reset if max or on writes
    else txdiv<=txdiv+1;

reg [8:0]txsh=10'h1FF;  // 9-bit shift register
always @(posedge clk)
    if (txwr) begin 
          `ifdef SIMULATION
          $write ("%c",pl[13:6]); $fflush ( );
          `endif
          txsh<={txdata,1'b0}; // Load data plus START bit
    end     
    else if (txdivmax) txsh<={1'b1,txsh[8:1]}; // shift 1'b1 in for future STOP bit
assign txd = txsh[0];

reg [3:0]txbitcnt=0;    // Bit counter (0000 means idle)
wire utxbusy=|txbitcnt;
always @(posedge clk) 
	if (uwr) txbitcnt<=4'd10; // start+data+stop = 10bits
	else if (utxbusy&txdivmax) txbitcnt<=txbitcnt-1;



endmodule

